from abc import ABC, abstractmethod
import numpy as np
from typing import Any, Dict, Mapping, Optional, Type, Union, List, TypeVar, Tuple
import zfit
import zfit.core
from hepstats.hypotests import Discovery
from hepstats.hypotests.calculators import AsymptoticCalculator, FrequentistCalculator
from hepstats.hypotests import UpperLimit
from hepstats.hypotests.parameters import POI, POIarray
from .utilities import PathType, write_json
from .Models import Model

__all__ = ["F", "BaseFitter"]

F = TypeVar("F", bound="BaseFitter")


class BaseFitter(ABC):
    def __init__(
        self,
        model: Model,
        retry: int = 10,
        minimizer: str = "Minuit",
        minimizer_args: Dict[Any, Any] = {"gradient": True},
        global_optimizer: str = "",
        global_optimizer_args: Dict[Any, Any] = {},
        truncated_limits: Optional[
            Union[Tuple[float, float], List[Tuple[float, float]]]
        ] = None,
    ) -> None:
        self.model = model
        self.retry = retry
        self.global_optimizer = global_optimizer
        self.global_optimizer_args = global_optimizer_args
        self.minimizer = getattr(zfit.minimize, minimizer)(**minimizer_args)
        self.truncated_limits = truncated_limits

        self.result_text = []  # type: List[str]
        self.nll: Optional[zfit.core.loss.BaseLoss] = None

    @property
    @abstractmethod
    def data(self) -> Union[zfit.Data, zfit.data.BinnedData]:
        pass

    @property
    @abstractmethod
    def lossfunction(self) -> Type[zfit.core.loss.BaseLoss]:
        pass

    @property
    @abstractmethod
    def eval_model(
        self,
    ) -> Union[zfit.core.basepdf.BasePDF, zfit.core.interfaces.ZfitPDF]:
        pass

    @property
    @abstractmethod
    def gof(self) -> float:
        pass

    def _calc_gof(self, counts: np.ndarray, centers: np.ndarray) -> float:
        limits = self.model.obs.limits
        low, high = np.min(limits), np.max(limits)

        assert self.model.model is not None
        counts_model = self.model.model.pdf(centers)
        counts_model_normed = (
            counts_model * np.sum(counts) * (float(high) - float(low)) / len(counts)
        )

        nPars = 0
        for parameter in self.model.parameters.values():
            if parameter.floating:
                nPars += 1

        gof = np.sum((counts - counts_model_normed) ** 2 / counts_model_normed) / (
            len(counts) - nPars
        )

        return gof

    def _save_result(self, params: Mapping[Any, Any]) -> None:
        floating_params = []

        for param in params.keys():
            floating_params.append(param.name)
            fitparameter = self.model.get_fitparameter(str(param.name))
            name = fitparameter.latex_name
            unit = fitparameter.unit
            value = float(params[param.name]["value"])
            lower = float(params[param.name]["errors"]["lower"])
            upper = float(params[param.name]["errors"]["upper"])

            self.model.update_fitparameter(
                str(param.name), value, value + lower, value + upper
            )

            self.result_text.append(
                "{name} = ${value:.3{c1}}^{{+{upper:.3{c2}}}}_{{{lower:.3{c3}}}}$ {unit}".format(
                    name=name,
                    value=value,
                    lower=lower,
                    upper=upper,
                    unit=unit,
                    c1="e" if (abs(value) > 1e3 or abs(value) < 1e-3) else "f",
                    c2="e" if (abs(upper) > 1e3 or abs(upper) < 1e-3) else "f",
                    c3="e" if (abs(lower) > 1e3 or abs(lower) < 1e-3) else "f",
                )
            )

        for parameterName in self.model.parameters.keys():
            if parameterName not in floating_params:
                fitparameter = self.model.parameters[parameterName]
                self.result_text.append(
                    "{name} = {value:.3{c1}} {unit} (fixed)".format(
                        name=fitparameter.latex_name,
                        value=fitparameter.value,
                        unit=fitparameter.unit,
                        c1="e"
                        if (
                            abs(fitparameter.value) > 1e3
                            or abs(fitparameter.value) < 1e-3
                        )
                        else "f",
                    )
                )

    def _randomize_parameters(self, model: Model) -> None:
        for param in model.parameters.values():
            modelparameter = model.get_modelparameter(param.name)
            if param.floating and param.shuffle:
                modelparameter.randomize(
                    minval=param.lower, maxval=param.upper, sampler=np.random.uniform
                )
            elif param.floating and param.shuffle and param.extended:
                modelparameter.set_value(param.value)

    def _single_fit(self) -> None:
        assert isinstance(self.eval_model, zfit.core.interfaces.ZfitPDF)
        nll = self.lossfunction(
            model=self.eval_model,
            data=self.data,
            constraints=self.model.constraints,
        )
        self.nll = nll
        if self.global_optimizer:
            minimizer_global = getattr(zfit.minimize, self.global_optimizer)(
                **self.global_optimizer_args
            )
            result_global = minimizer_global.minimize(nll).update_params()
            result_global.errors()
            result_global.hesse()
            print("Done with global minimization")
            print(result_global.params)
            nll = result_global.loss

        minimizer_local = self.minimizer
        result = minimizer_local.minimize(nll)
        result = result.update_params()
        # calculate errors
        result.errors()
        try:
            result.hesse()
        except ValueError:
            pass
        self.zfitresult = result
        # Information on all the parameters in the fit
        params = result.params
        try:
            covariance = result.covariance(as_dict=True)
            assert isinstance(covariance, dict)
            self.model.covariance = covariance
        except ValueError:
            pass
        self.model.gof = self.gof
        print("Done with local minimization")
        print(params)

        self._save_result(params=params)

    def _fit(
        self,
        model: Model,
        shuffle_params: Union[None, List[str]] = None,
        counter: int = 0,
    ) -> None:
        try:
            self._single_fit()

        except BaseException as exception:
            print(f"[Fitter]: Fit attempt {counter + 1} failed!")
            print(exception)

            if counter > self.retry:
                raise BaseException("[Fitter]: Too much Fit retries.") from exception

            self._randomize_parameters(model=model)

            counter += 1
            self._fit(model=model, shuffle_params=shuffle_params, counter=counter)

    def fit(self) -> None:
        self._fit(model=self.model)

    def get_calculator(
        self,
        calculator: str = "AsymptoticCalculator",
        calculator_kwargs: Dict[str, Any] = {"asimov_bins": 100},
    ) -> Union[AsymptoticCalculator, FrequentistCalculator]:
        assert self.nll is not None, "No NLL yet calculated, please check the fit"
        if calculator == "AsymptoticCalculator":
            return AsymptoticCalculator(self.nll, self.minimizer, **calculator_kwargs)
        elif calculator == "FrequentistCalculator":
            return FrequentistCalculator(self.nll, self.minimizer, **calculator_kwargs)
        else:
            raise ValueError("Calculator not recognized")

    def calculate_significance(
        self,
        parameter_name: str,
        parameter_range: np.ndarray,
        calculator: str = "AsymptoticCalculator",
        calculator_kwargs: Dict[str, Any] = {"asimov_bins": 100},
    ) -> Dict[str, float]:
        parameter = self.model.get_modelparameter(parameter_name)
        calculator = self.get_calculator(
            calculator=calculator, calculator_kwargs=calculator_kwargs
        )
        calculator.bestfit = self.zfitresult
        poinull = POIarray(parameter, parameter_range)
        discovery_test = Discovery(calculator, poinull)
        pnull, significance = discovery_test.result()
        result = {"pvalue": pnull, "significance": significance}
        self.model.significance = result

        return result

    def calculate_upperlimit(
        self,
        parameter_name: str,
        parameter_range: np.ndarray,
        alt_value: float = 0.0,
        alpha: float = 0.05,
        calculator: str = "AsymptoticCalculator",
        calculator_kwargs: Dict[str, Any] = {"asimov_bins": 100},
    ) -> Dict[str, float]:
        parameter = self.model.get_modelparameter(parameter_name)
        calculator = self.get_calculator(
            calculator=calculator, calculator_kwargs=calculator_kwargs
        )
        calculator.bestfit = self.zfitresult
        poinull = POIarray(parameter, parameter_range)
        poialt = POI(parameter, alt_value)
        ul = UpperLimit(calculator, poinull, poialt)
        result = ul.upperlimit(alpha=alpha, CLs=True)
        self.model.upperlimit = result

        return result

    def write_result(self, savepath: PathType) -> None:
        write_json(array=self.model.dict, savepath=savepath)
