from abc import ABC, abstractmethod
from typing import Dict, Optional, Union, List
import zfit
import zfit.core

from .FitParameter import (
    Fitparameter,
    FitCovarianceMatrix,
    CovarianceInputType,
    CovarianceType,
)


class BaseModel(ABC):
    def __init__(
        self,
        obs: Union[zfit.core.space.Space, zfit.core.space.MultiSpace],
    ) -> None:
        self.obs = obs

        self.name = "Model"
        self._parameters: Dict[str, Fitparameter] = {}
        self._modelparameters: Dict[
            str, Union[zfit.Parameter, zfit.ComposedParameter]
        ] = {}
        self._covariance: CovarianceType = {}
        self._model: Optional[zfit.core.basepdf.BasePDF] = None
        self._gof: float = 0.0
        self._significance: Dict[str, float] = {}
        self._upperlimit: Dict[str, float] = {}
        self._constraints: List[zfit.core.constraint.TFProbabilityConstraint] = []

    def __str__(self) -> str:
        string = "Model Wrapper"
        string += f"\nParameters:\n{self.parameters}"
        string += f"\nModel:\n{self.name}"

        return string

    @property
    def parameters(self) -> Dict[str, Fitparameter]:
        return self._parameters

    @property
    def constraints(
        self,
    ) -> Optional[List[zfit.core.constraint.TFProbabilityConstraint]]:
        return self._constraints if self._constraints else None

    @property
    def modelparameters(
        self,
    ) -> Dict[str, Union[zfit.Parameter, zfit.ComposedParameter]]:
        if not self.parameters or self._modelparameters:
            return self._modelparameters

        for key, parameter in self.parameters.items():
            if key not in self._modelparameters:
                self._modelparameters[key] = parameter.to_zfitParameter

                if parameter.composed:
                    self._modelparameters.update(parameter.sub_parameter)

        return self._modelparameters

    @property
    def covariance(self) -> CovarianceType:
        return self._covariance

    @covariance.setter
    def covariance(self, covariance: CovarianceInputType) -> None:
        fitCovariance = FitCovarianceMatrix(covariance)
        self._covariance.update(fitCovariance.covariance)

    @property
    @abstractmethod
    def model(self) -> Optional[zfit.core.basepdf.BasePDF]:
        pass

    @property
    def gof(self) -> float:
        return self._gof

    @gof.setter
    def gof(self, gof: float) -> None:
        self._gof = gof

    @property
    def significance(self) -> Dict[str, float]:
        return self._significance

    @significance.setter
    def significance(self, significance: Dict[str, float]) -> None:
        self._significance = significance

    @property
    def upperlimit(self) -> Dict[str, float]:
        return self._upperlimit

    @upperlimit.setter
    def upperlimit(self, upperlimit: Dict[str, float]) -> None:
        self._upperlimit = upperlimit

    def get_fitparameter(self, name: str) -> Fitparameter:
        OutputParameter = Fitparameter(name)
        if self.parameters:
            for parameter in self.parameters.values():
                if name == parameter.name:
                    OutputParameter = parameter

        return OutputParameter

    def update_fitparameter(
        self, name: str, value: float, lower: float, upper: float
    ) -> None:
        if self.parameters:
            for parameter in self.parameters.values():
                if name == parameter.name:
                    parameter.value = value
                    parameter.lower = lower
                    parameter.upper = upper

    def get_modelparameter(
        self, name: str
    ) -> Union[zfit.Parameter, zfit.ComposedParameter]:
        OutputParameter = None
        for parameter in self.modelparameters.values():
            if name == str(parameter.name):
                OutputParameter = parameter

        assert OutputParameter is not None
        return OutputParameter

    def update_modelparameter(self, name: str, value: float) -> None:
        if self.modelparameters:
            for parameter in self.modelparameters.values():
                if name == str(parameter.name):
                    parameter.set_value(value)

    def get_fraction(
        self, parametername: str = ""
    ) -> Optional[Union[zfit.Parameter, zfit.ComposedParameter]]:
        if parametername:
            return self.get_modelparameter(parametername)

        else:
            for parametername, parameter in self.parameters.items():
                if parameter.fraction:
                    return self.get_modelparameter(parametername)

            else:
                raise KeyError("No fraction parameter can be found")

    def add_constraint(
        self, parameter: str, constraint: str = "GaussianConstraint", **kwargs
    ) -> None:
        param = self.get_modelparameter(parameter)
        constraint_class = getattr(zfit.core.constraint, constraint)
        self._constraints.append(constraint_class(param, **kwargs))
