from typing import Any, Dict, Optional, Tuple, Union, List
import numpy as np

import zfit

__all__ = [
    "Fitparameter",
    "dictToFitparameter",
    "FitCovarianceMatrix",
    "CovarianceInputType",
    "CovarianceType",
]


CovarianceInputType = Dict[Tuple[zfit.Parameter, zfit.Parameter], Optional[np.float64]]
CovarianceValueType = Dict[str, Optional[float]]
CovarianceType = Dict[str, CovarianceValueType]


class Fitparameter:
    def __init__(
        self,
        name: str,
        value: float = 1.0,
        lower: Optional[float] = None,
        upper: Optional[float] = None,
        step_size: Optional[float] = None,
        floating: bool = True,
        latex_name: Optional[str] = None,
        unit: str = "",
        shuffle: bool = True,
        extended: bool = False,
        composed: bool = False,
        fraction: bool = False,
        func: str = "",
        params: Dict[str, Any] = {},
    ) -> None:
        self.name = name
        self.value = value
        self.lower = lower
        self.upper = upper
        self.step_size = step_size
        self.floating = floating
        self.latex_name = name if latex_name is None else latex_name
        self.unit = unit
        self.shuffle = shuffle
        self.extended = extended
        self.composed = composed
        self.fraction = fraction
        self.func = func
        self.params = params

        self.sub_parameter: Dict[str, zfit.Parameter] = {}

    def __str__(self) -> str:
        output = ""
        output += f"Name: {self.name}\n"
        output += f"Value: {self.value}\n"
        output += f"Lower: {self.lower}\n"
        output += f"Upper: {self.upper}\n"
        output += f"Step Size: {self.step_size}\n"
        output += f"Floating: {self.floating}\n"
        output += f"Latex Name: {self.latex_name}\n"
        output += f"Unit: {self.unit}\n"
        output += f"Shuffle: {self.shuffle}\n"
        output += f"Extended: {self.extended}\n"
        output += f"Composed: {self.composed}\n"
        output += f"Fraction: {self.fraction}\n"
        output += f"Func: {self.func}\n"
        output += f"Params: {self.params}\n"
        return output

    @property
    def parameter_info_dict(self) -> dict:
        return {
            "name": self.name,
            "value": self.value,
            "lower": self.lower,
            "upper": self.upper,
            "step_size": self.step_size,
            "floating": self.floating,
            "latex_name": self.latex_name,
            "unit": self.unit,
            "shuffle": self.shuffle,
            "extended": self.extended,
            "composed": self.composed,
            "fraction": self.fraction,
            "func": self.func,
            "params": self.params,
        }

    @property
    def parameter_zfit_dict(self) -> dict:
        return {
            "name": self.name,
            "value": self.value,
            "lower": self.lower,
            "upper": self.upper,
            "step_size": self.step_size,
            "floating": self.floating,
        }

    @property
    def to_zfitParameter(self) -> Union[zfit.Parameter, zfit.ComposedParameter]:
        if not self.composed:
            return zfit.Parameter(**self.parameter_zfit_dict)
        else:
            assert self.func, "Cannot build composed parameter without function."
            assert self.params, "Cannot build composed parameter without parameters"

            func = eval(self.func)
            parameter_dict = {}
            for parametername, parametervalues in self.params.items():
                parameter_dict.update(
                    {parametername: Fitparameter(**parametervalues).to_zfitParameter}
                )

            self.sub_parameter = parameter_dict

            return zfit.ComposedParameter(self.name, func, params=parameter_dict)


class FitCovarianceElement:
    def __init__(
        self,
        parameter1: zfit.Parameter,
        parameter2: zfit.Parameter,
        value: Optional[np.float64],
    ) -> None:
        self.parameter1 = parameter1.name
        self.parameter2 = parameter2.name
        self.value = value if value is None else float(value)

    @property
    def dict(self) -> CovarianceType:
        return {self.parameter1: {self.parameter2: self.value}}


class FitCovarianceMatrix:
    def __init__(self, covariance: CovarianceInputType) -> None:
        self._zfitcovariance = covariance
        self._covarianceElementList: List[FitCovarianceElement] = []
        self._covarianceMatrix: CovarianceType = {}

    def _read_covariance(self) -> None:
        for (parameter1, parameter2), value in self._zfitcovariance.items():
            covarianceElement = FitCovarianceElement(
                parameter1=parameter1, parameter2=parameter2, value=value
            )
            self._covarianceElementList.append(covarianceElement)

    @staticmethod
    def __merge_dicts(
        dict1: CovarianceType,
        dict2: CovarianceType,
    ) -> None:
        for key, value in dict2.items():
            if key in dict1.keys():
                dict1[key].update(value)
            else:
                dict1.update({key: value})

    def _sort_elements(self) -> CovarianceType:
        covarianceMatrix = {}  # type: CovarianceType
        for covarianceElement in self._covarianceElementList:
            FitCovarianceMatrix.__merge_dicts(covarianceMatrix, covarianceElement.dict)

        return covarianceMatrix

    @property
    def covariance(self) -> CovarianceType:
        if not self._covarianceMatrix:
            self._read_covariance()
            self._covarianceMatrix = self._sort_elements()

        return self._covarianceMatrix


def dictToFitparameter(**kwargs: Any) -> Dict[str, Fitparameter]:
    fitparameterdict = {}

    for key, value in kwargs.items():
        if not isinstance(value, list):
            fitparameterdict[key] = Fitparameter(**value)
            if "composed" in value and value["composed"]:
                assert "params" in value
                for param_key, param_value in value["params"].items():
                    fitparameterdict[param_key] = Fitparameter(**param_value)

        else:
            for val in value:
                fitparameterdict[val["name"]] = Fitparameter(**val)

    return fitparameterdict
