from typing import Optional, Type, Union
import numpy as np
import zfit
import zfit.core

from .BaseFitter import BaseFitter

__all__ = ["Fitter", "HistFitter"]


class Fitter(BaseFitter):
    def __init__(
        self,
        array: np.ndarray,
        weights: Optional[np.ndarray] = None,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)

        self.array = array
        self.weights = weights
        self.zfitresult = None

    @property
    def data(self) -> zfit.Data:
        assert self.model.model is not None
        obs = self.model.model.space
        data = zfit.Data.from_numpy(obs=obs, array=self.array, weights=self.weights)

        return data

    @property
    def lossfunction(self) -> Type[zfit.core.loss.BaseLoss]:
        assert self.model.model is not None
        if self.model.model.is_extended:
            lossfunction = zfit.loss.ExtendedUnbinnedNLL
        else:
            lossfunction = zfit.loss.UnbinnedNLL

        return lossfunction

    @property
    def eval_model(
        self,
    ) -> Union[zfit.core.basepdf.BasePDF, zfit.core.interfaces.ZfitPDF]:
        assert self.model.model is not None
        pdf = self.model.model
        if self.truncated_limits:
            if len(self.truncated_limits) > 1:
                assert isinstance(self.truncated_limits, list)
                new_range = [
                    zfit.Space(
                        obs=self.model.obs,
                        lower=limits[0],
                        upper=limits[1],
                    )
                    for limits in self.truncated_limits
                ]
                pdf = self.model.model.to_truncated(
                    limits=new_range, extended=self.model.model.is_extended
                )
            else:
                pdf = self.model.model.to_truncated(
                    limits=self.truncated_limits, extended=self.model.model.is_extended
                )
        return pdf

    @property
    def gof(self) -> float:
        limits = self.model.obs.limits
        low, high = np.min(limits), np.max(limits)
        bins = np.linspace(float(low), float(high), 100)
        counts_array, edges = np.histogram(self.array, bins=bins, weights=self.weights)
        bin_centers = (edges[1:] + edges[:-1]) / 2

        gof = self._calc_gof(counts=counts_array, centers=bin_centers)

        return float(gof)


class HistFitter(BaseFitter):
    def __init__(self, counts: np.ndarray, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.counts = counts

    @property
    def obs(self) -> Union[zfit.core.space.Space, zfit.core.space.MultiSpace]:
        assert self.model.model is not None
        obs_unbin = self.model.model.space

        assert obs_unbin.obs is not None
        binning = zfit.binned.RegularBinning(
            name=obs_unbin.obs[0],
            bins=len(self.counts),
            start=np.min(np.array(obs_unbin.limits)),
            stop=np.max(np.array(obs_unbin.limits)),
        )

        obs = zfit.Space(obs_unbin.obs[0], binning=binning)

        return obs

    @property
    def data(self) -> zfit.data.BinnedData:
        data = zfit.data.BinnedData.from_tensor(
            space=self.obs, values=self.counts, variances=True
        )

        return data

    @property
    def lossfunction(self) -> Type[zfit.core.loss.BaseLoss]:
        assert self.model.model is not None
        if self.model.model.is_extended:
            lossfunction = zfit.loss.ExtendedBinnedNLL
        else:
            lossfunction = zfit.loss.BinnedNLL

        return lossfunction

    @property
    def eval_model(
        self,
    ) -> Union[zfit.core.basepdf.BasePDF, zfit.core.interfaces.ZfitPDF]:
        assert self.model.model is not None
        assert self.model.model is not None
        pdf = zfit.pdf.BinnedFromUnbinnedPDF(self.model.model, self.obs)
        if self.truncated_limits:
            pdf = zfit.pdf.TruncatedPDF(
                zfit.pdf.BinnedFromUnbinnedPDF(self.model.model, self.obs),
                (
                    zfit.Space(obs=self.model.obs, limits=limits)
                    for limits in self.truncated_limits
                ),
                self.model.obs,
            )
        return pdf

    @property
    def gof(self) -> float:
        assert self.obs.binning is not None
        bin_centers = self.obs.binning.centers[0]

        gof = self._calc_gof(counts=self.counts, centers=bin_centers)

        return float(gof)
