import enum
import re
from typing import List, NamedTuple, Optional, Type

import zfit.models.functor


class TokenType(enum.Enum):
    T_MODEL = 0
    T_NUM = 1
    T_PLUS = 2
    T_MULT = 3
    T_CONV = 4
    T_LPAR = 5
    T_RPAR = 6
    T_END = 7
    T_EXT = 8


class Token(NamedTuple):
    type: TokenType
    operation: Optional[Type[zfit.models.functor.BaseFunctor]] = None


class Node:
    def __init__(
        self,
        token_type: TokenType,
        value: Optional[str] = None,
        ext_param: Optional[str] = None,
    ):
        self.token_type = token_type
        self.value = value
        self.ext_param = ext_param  # For extended models, stores the parameter
        self.children = []  # type: List[Node]


KnownMappings = {
    "+": Token(type=TokenType.T_PLUS, operation=zfit.pdf.SumPDF),
    "*": Token(type=TokenType.T_MULT, operation=zfit.pdf.ProductPDF),
    "**": Token(type=TokenType.T_CONV, operation=zfit.pdf.FFTConvPDFV1),
    "(": Token(type=TokenType.T_LPAR),
    ")": Token(type=TokenType.T_RPAR),
}


class ModelParser:
    def __init__(self, modelstring: str) -> None:
        self.modelstring = modelstring.replace(" ", "")  # type: str
        self._nodeTokens = []  # type: List[Node]

    @staticmethod
    def __match(tokens: List[Node], token: TokenType) -> Node:
        if tokens[0].token_type == token:
            return tokens.pop(0)
        else:
            raise Exception("Invalid syntax on token {}".format(tokens[0].token_type))

    @staticmethod
    def __baseParse(tokens: List[Node]) -> Node:
        left_node = ModelParser.__priorParse(tokens)

        while tokens[0].token_type in [TokenType.T_PLUS, TokenType.T_CONV]:
            node = tokens.pop(0)
            node.children.append(left_node)
            node.children.append(ModelParser.__priorParse(tokens))
            left_node = node

        return left_node

    @staticmethod
    def __priorParse(tokens: List[Node]) -> Node:
        left_node = ModelParser.__parenthesesParse(tokens)

        while tokens[0].token_type in [TokenType.T_MULT]:
            node = tokens.pop(0)
            node.children.append(left_node)
            node.children.append(ModelParser.__parenthesesParse(tokens))
            left_node = node

        return left_node

    @staticmethod
    def __parenthesesParse(tokens: List[Node]) -> Node:
        if tokens[0].token_type == TokenType.T_MODEL:
            return tokens.pop(0)
        if tokens[0].token_type == TokenType.T_EXT:
            return tokens.pop(0)

        ModelParser.__match(tokens, TokenType.T_LPAR)
        expression = ModelParser.__baseParse(tokens)
        ModelParser.__match(tokens, TokenType.T_RPAR)

        return expression

    def _lexical_analysis(self, input_string: str) -> List[Node]:
        token_pattern = r"(\w+@\w+|\w+@\([^\)]+\)|\w+|[+*()])"
        tokens = re.findall(token_pattern, input_string)

        for expr in tokens:
            if re.match(r"\w+@(\w+|\()", expr):
                param_name, model_name = expr.split("@")
                token = Node(TokenType.T_EXT, value=model_name, ext_param=param_name)
                self._nodeTokens.append(token)

            elif re.match(r"\w+", expr):
                token = Node(TokenType.T_MODEL, value=expr)
                self._nodeTokens.append(token)
            elif expr in KnownMappings.keys():
                token_type = KnownMappings[expr]
                token = Node(token_type.type, value=expr)
                self._nodeTokens.append(token)
            elif expr == "@":
                continue  # '@' is handled as part of the extended model token
            else:
                raise Exception("Invalid token: {}".format(expr))

        return self._nodeTokens

    def parse(self) -> Node:
        tokens = self._lexical_analysis(self.modelstring)
        self._nodeTokens.append(Node(TokenType.T_END))

        fNode = self.__baseParse(tokens)
        self.__match(tokens, TokenType.T_END)

        return fNode
