import copy
from typing import Any, Dict, List, Optional, Union
import zfit
import zfit.core

from .FitParameter import dictToFitparameter, CovarianceType
from .BaseModel import BaseModel
from .ModelParser import ModelParser, Node, TokenType, KnownMappings
from .utilities import set_obs

zfit.settings.changed_warnings.new_sum = False

__all__ = ["ModelsReadType", "LimitType", "Model"]

ModelsReadType = Dict[str, Dict[str, Dict[str, Dict[str, Any]]]]
LimitType = Union[List[float], List[List[float]], list]
ModelsOutputType = Union[
    ModelsReadType, str, LimitType, CovarianceType, float, Dict[str, float]
]


class Model(BaseModel):
    def __init__(
        self,
        obs: Union[zfit.core.space.Space, zfit.core.space.MultiSpace],
        modelstring: str,
        models: ModelsReadType,
        modelname: Optional[str] = None,
    ) -> None:
        super().__init__(obs)

        self.name = modelname if modelname is not None else "Model"
        self._modelstring = modelstring
        self._constraint_info: Dict[str, Dict[str, Any]] = {}
        self._models = self._prepare_models(copy.deepcopy(models))
        for constraint in self._constraint_info.values():
            self.add_constraint(**constraint)
        parser = ModelParser(modelstring)
        self._node = parser.parse()

    def _prepare_models(self, models: ModelsReadType) -> Dict[str, Dict[str, Any]]:
        tmpmodels = {}  # type: Dict[str, Dict[str, Dict[str, Any]]]

        for modelname, modelparameters in models.items():
            if "constraint" not in modelname:
                assert modelname in self._modelstring, (
                    f"[Models]: Given model '{modelname}' not in given modelstring '{self._modelstring}'"
                )

                assert "pdf" in modelparameters.keys(), (
                    f"[Models]: Please provide a pdf for model '{modelname}'"
                )
                assert "parameter" in modelparameters.keys(), (
                    f"[Models]: Please provide modelparameters for model '{modelname}'"
                )

                for parametername, values in modelparameters["parameter"].items():
                    try:
                        parameter = {values["name"]: values}
                    except TypeError:
                        parameter = {parametername: values}

                    self._parameters.update(dictToFitparameter(**parameter))

                tmpmodels.update({modelname: modelparameters})

            elif "constraint" in modelname:
                self._constraint_info.update({modelname: modelparameters})
            else:
                raise KeyError(f"Cannot interpret input from {modelname}")

        return tmpmodels

    def _compute_model(self, node: Node) -> zfit.core.basepdf.BasePDF:
        if node.token_type == TokenType.T_EXT:
            assert node.value is not None
            assert node.ext_param is not None

            parser = ModelParser(node.value)
            model = self._compute_model(parser.parse())
            ext_param = self.modelparameters[node.ext_param]
            zfit_model = model.create_extended(ext_param)
            return zfit_model

        elif node.token_type == TokenType.T_MODEL:
            assert node.value is not None

            model = self._models[node.value]
            pdf = model["pdf"]
            kwargs = model.get("kwargs", {})

            parameter = {}  # type: Dict[str, zfit.Parameter]

            for parametername, values in model["parameter"].items():
                if not isinstance(values, list):
                    if (
                        not self.parameters[values["name"]].fraction
                        and not self.parameters[values["name"]].extended
                    ):
                        parameter.update(
                            {parametername: self.modelparameters[values["name"]]}
                        )

                else:
                    if len(values) > 0:
                        parameter.update(
                            {
                                parametername: [
                                    self.modelparameters[value["name"]]
                                    for value in values
                                ]
                            }
                        )
                    else:
                        parameter.update({parametername: []})

            obs = self.obs
            if "limits" in model.keys():
                limits = model["limits"]
                assert isinstance(limits, tuple) or isinstance(limits, list)
                assert self.obs.obs is not None
                obs = set_obs(
                    self.obs.obs[0],
                    (tuple(limits) if isinstance(limits[0], float) else limits),
                )

            zfit_model = getattr(zfit.pdf, pdf)(obs=obs, **parameter, **kwargs)

            return zfit_model

        else:
            left_result = self._compute_model(node.children[0])
            right_result = self._compute_model(node.children[1])

            operation = None
            for token in KnownMappings.values():
                if node.token_type == token.type:
                    operation = token.operation

            assert operation is not None
            op_kwargs = {}  # type: Dict[str, Any]

            if node.token_type == TokenType.T_PLUS:
                if left_result.is_extended != right_result.is_extended:
                    raise Exception(
                        "[Models]: SumPDF does not work with one pdf extended and not the other."
                    )

                if not left_result.is_extended:
                    assert node.children[0].value is not None
                    model = self._models[node.children[0].value]

                    # find the parameter that steers the fraction of the sum
                    fraction_parameter_name = ""
                    for modelparameter in model["parameter"].values():
                        if "fraction" in modelparameter.keys():
                            if modelparameter["fraction"]:
                                fraction_parameter_name = modelparameter["name"]
                                break
                    assert fraction_parameter_name, "No fraction parameter found"
                    fraction_parameter = self.parameters.get(
                        fraction_parameter_name, None
                    )
                    assert fraction_parameter, "Fraction parameter not found"

                    op_kwargs.update({"fracs": fraction_parameter.to_zfitParameter})

                op_kwargs.update({"pdfs": [left_result, right_result]})

            elif node.token_type == TokenType.T_CONV:
                op_kwargs.update({"func": left_result})
                op_kwargs.update({"kernel": right_result})
                op_kwargs.update({"n": int(1e3)})
            elif node.token_type == TokenType.T_MULT:
                op_kwargs.update({"pdfs": [left_result, right_result]})
            else:
                raise Exception("[Models]: Something went terribly wrong")

            return operation(**op_kwargs)

    @property
    def model(self) -> Optional[zfit.core.basepdf.BasePDF]:
        if self._model is None:
            self._model = self._compute_model(self._node)

        return self._model

    @property
    def models(self) -> List["Model"]:
        models: List["Model"] = []
        for modelname, model in self._models.items():
            models.append(
                Model(
                    obs=self.obs,
                    modelstring=modelname,
                    models={modelname: model},
                    modelname=modelname,
                )
            )

        return models

    @property
    def dict(self) -> Dict[str, ModelsOutputType]:
        output = {
            "modelname": self.name,
            "modelstring": self._modelstring,
        }  # type: Dict[str, ModelsOutputType]
        modelparameter = {}  # type: ModelsReadType
        for submodelname, submodelparameter in self._models.items():
            # get update parameter from model
            parameter = {}  # type: Dict[str, Dict[str, Any]]
            for parametername, values in submodelparameter["parameter"].items():
                if not isinstance(values, list):
                    parameter.update(
                        {
                            parametername: self.parameters[
                                values["name"]
                            ].parameter_info_dict
                        }
                    )
                else:
                    parameter.update(
                        {
                            parametername: [
                                self.parameters[value["name"]].parameter_info_dict
                                for value in values
                            ]
                        }
                    )
            submodelparameter.pop("parameter")
            submodelparameter.update({"parameter": parameter})
            modelparameter.update({submodelname: submodelparameter})

        modelparameter.update(self._constraint_info)
        output.update({"modelparameter": modelparameter})
        output.update({"covariance": self.covariance})

        limits = []  # type: LimitType
        try:
            assert self.obs.rect_limits_np is not None
            limits = [
                float(self.obs.rect_limits_np[0]),
                float(self.obs.rect_limits_np[1]),
            ]

        except zfit.exception.MultipleLimitsNotImplemented:
            for space in self.obs.__dict__["spaces"]:
                assert space.rect_limits_np is not None
                tmplimits = [
                    float(space.rect_limits_np[0]),
                    float(space.rect_limits_np[1]),
                ]
                limits.append(tmplimits)  # type: ignore

        output.update({"limits": limits})
        output.update({"gof": self.gof})
        output.update({"significance": self.significance})
        output.update({"upperlimit": self.upperlimit})

        return output
