import os
from typing import Any, Dict, Union, List, AnyStr

import numpy as np
import pandas as pd
import zfit
import zfit.core
import json

__all__ = [
    "PathType",
    "OutputType",
    "set_obs",
    "clear_existing_parameters",
    "write_json",
    "read_json",
]

PathType = Union[str, AnyStr, os.PathLike]
OutputType = Dict[str, Any]


def set_obs(
    variable: str, tuples: Union[List[tuple], tuple]
) -> Union[zfit.core.space.Space, zfit.core.space.MultiSpace]:
    """Build zfit Space for given tuple(s)

    Args:
        variable (str): name of space
        tuples (Union[List[tuple], tuple]): range(s) for given space

    Returns:
        zfit.util.ztyping.SpaceType: zfit space
    """

    if isinstance(tuples, list):
        obs = zfit.Space(variable, tuples[0])
        for limits in tuples[1:]:
            obs = obs + zfit.Space(variable, limits)

    else:
        obs = zfit.Space(variable, tuples)

    return obs


def clear_existing_parameters() -> None:
    """Helper function to clear existing zfit parameters in fits with jupyter notebooks

    Args:
        parameters (List[str]): Parameter names of zfit parameter
    """

    if len(zfit.Parameter._existing_params) != 0:
        for param in list(zfit.Parameter._existing_params.keys()):
            del zfit.Parameter._existing_params[param]


def write_json(
    array: Union[np.ndarray, list, dict, pd.DataFrame], savepath: PathType
) -> PathType:
    if isinstance(array, dict):
        with open(savepath, "w") as f:
            json.dump(array, f, indent=4)
    else:
        if not isinstance(array, pd.DataFrame):
            df = pd.DataFrame(array)

        else:
            df = array

        savedir = ""
        for part in str(savepath).split("/")[:-1]:
            savedir += str(part) + "/"
        if not os.path.exists(savedir):
            os.makedirs(savedir)

        df.to_json(savepath)

    return savepath


def read_json(filename: PathType) -> OutputType:
    with open(filename, "r") as f:
        json_dict = json.load(f)

    return json_dict
